@@ -5,7 +5,7 @@
 // Created: Mon Sep  7 10:18:28 2009
 //
 // Author(s): Atgeirr F Rasmussen <atgeirr@sintef.no>
-//            Brd Skaflestad     <bard.skaflestad@sintef.no>
+//            Bård Skaflestad     <bard.skaflestad@sintef.no>
 //
 // $Date$
 //
@@ -41,7 +41,7 @@
 #include <dune/istl/owneroverlapcopy.hh>
 #include "GridPartitioning.hpp"
 #include <opm/grid/CpGrid.hpp>
-#include <opm/grid/cpgrid/CpGridData.hpp>
+#include <opm/grid/cpgrid/CpGridDataTraits.hpp>
 #include <opm/grid/common/ZoltanPartition.hpp>
 #include <stack>
 
@@ -204,7 +204,10 @@ namespace Dune
         const coord_t& lc_size = grid.logicalCartesianSize();
         for (int i = 0; i < 3; ++i) {
             if (initial_split[i] > lc_size[i]) {
-                OPM_THROW(std::runtime_error, "In direction " << i << " requested splitting " << initial_split[i] << " size " << lc_size[i]);
+                OPM_THROW(std::runtime_error,
+                          "In direction "  + std::to_string(i) +
+                          " requested splitting " + std::to_string(initial_split[i]) +
+                          " size " + std::to_string(lc_size[i]));
             }
         }
 
@@ -290,7 +293,7 @@ void addOverlapCornerCell(const CpGrid& grid, int owner,
     //  I I O E         I I O E 
     //  O O O E         O O E E
     //  E E E E         E E E E
-    using AttributeSet = Dune::cpgrid::CpGridData::AttributeSet;
+    using AttributeSet = Dune::cpgrid::CpGridDataTraits::AttributeSet;
     const CpGrid::LeafIndexSet& ix = grid.leafIndexSet();
     int my_index = ix.index(from);
     int nb_index = ix.index(neighbor);
@@ -379,7 +382,7 @@ void addOverlapLayer(const CpGrid& grid, int index, const CpGrid::Codim<0>::Enti
                          std::vector<std::tuple<int,int,char>>& exportList,
                          bool addCornerCells, int recursion_deps)
     {
-        using AttributeSet = Dune::cpgrid::CpGridData::AttributeSet;
+        using AttributeSet = Dune::cpgrid::CpGridDataTraits::AttributeSet;
         const CpGrid::LeafIndexSet& ix = grid.leafIndexSet();
         for (CpGrid::LeafIntersectionIterator iit = e.ileafbegin(); iit != e.ileafend(); ++iit) {
             if ( iit->neighbor() ) {
@@ -469,11 +472,7 @@ void addOverlapLayer(const CpGrid& grid, int index, const CpGrid::Codim<0>::Enti
     int addOverlapLayer(const CpGrid& grid, const std::vector<int>& cell_part,
                         std::vector<std::tuple<int,int,char>>& exportList,
                         std::vector<std::tuple<int,int,char,int>>& importList,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                         const Communication<Dune::MPIHelper::MPICommunicator>& cc,
-#else
-                        const CollectiveCommunication<Dune::MPIHelper::MPICommunicator>& cc,
-#endif
                         [[maybe_unused]] bool addCornerCells,
                         [[maybe_unused]] const double* trans, int layers)
     {
