@@ -20,12 +20,14 @@
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
+#if HAVE_MPI // no code in this file without MPI. Skip includes-
 #include <opm/grid/common/ZoltanPartition.hpp>
-#include <opm/grid/utility/OpmParserIncludes.hpp>
+#include <opm/grid/utility/OpmWellType.hpp>
 #include <opm/grid/cpgrid/CpGridData.hpp>
 #include <opm/grid/cpgrid/Entity.hpp>
 #include <algorithm>
 #include <type_traits>
+#endif
 
 namespace Dune
 {
@@ -39,11 +41,7 @@ std::tuple<std::vector<int>, std::vector<std::pair<std::string,bool>>,
            std::vector<std::tuple<int,int,char,int> >,
            WellConnections>
 makeImportAndExportLists(const Dune::CpGrid& cpgrid,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                          const Dune::Communication<MPI_Comm>& cc,
-#else
-                         const Dune::CollectiveCommunication<MPI_Comm>& cc,
-#endif
                          const std::vector<Dune::cpgrid::OpmWellType> * wells,
                          const Dune::cpgrid::CombinedGridWellGraph* gridAndWells,
                          int root,
@@ -110,12 +108,15 @@ makeImportAndExportLists(const Dune::CpGrid& cpgrid,
                                                 gidGetter,
                                                 *wells,
                                                 gridAndWells->getWellConnections(),
+                                                gridAndWells->getWellsGraph(),
                                                 myExportList, myImportList,
                                                 cc);
 
 
 #ifndef NDEBUG
-            int index = 0;
+            std::size_t index = 0;
+            std::unordered_set<int> distributed_wells;
+
             for( auto well : gridAndWells->getWellsGraph() )
             {
                 int part=parts[index];
@@ -129,10 +130,19 @@ makeImportAndExportLists(const Dune::CpGrid& cpgrid,
                 }
                 if ( cells_on_other.size() )
                 {
-                    OPM_THROW(std::domain_error, "Well is distributed between processes, which should not be the case!");
+                    distributed_wells.insert(index);
                 }
                 ++index;
             }
+            auto num_dist_wells = cc.sum(distributed_wells.size());
+            if (num_dist_wells) {
+                OPM_THROW(std::domain_error,
+                          std::to_string(num_dist_wells) + " well"
+                              +  ((num_dist_wells >= 2) ? "s are" : " is")
+                              + " distributed between processes, which should not be the case!");
+            }
+
+
 #endif
         }
     }
@@ -153,11 +163,7 @@ template<class Id>
 std::tuple<int, std::vector<Id> >
 scatterExportInformation(int numExport, const Id* exportGlobalGids,
                          const int* exportToPart, int root,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                          const Communication<MPI_Comm>& cc)
-#else
-                         const Dune::CollectiveCommunication<MPI_Comm>& cc)
-#endif
 {
     int numImport;
     std::vector<int> numberOfExportedVerticesPerProcess;
@@ -233,11 +239,7 @@ std::tuple<std::vector<int>, std::vector<std::pair<std::string,bool>>,
            std::vector<std::tuple<int,int,char,int> >,
            WellConnections>
 makeImportAndExportLists(const Dune::CpGrid&,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                          const Communication<MPI_Comm>&,
-#else
-                         const Dune::CollectiveCommunication<MPI_Comm>&,
-#endif
                          const std::vector<Dune::cpgrid::OpmWellType>*,
                          const Dune::cpgrid::CombinedGridWellGraph*,
                          int,
@@ -253,11 +255,7 @@ template
 std::tuple<int, std::vector<int> >
 scatterExportInformation(int numExport, const int*,
                          const int*, int,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                          const Communication<MPI_Comm>&);
-#else
-                         const Dune::CollectiveCommunication<MPI_Comm>&);
-#endif
 #endif // HAVE_MPI
 } // end namespace Dune
 } // end namespace cpgrid
@@ -292,11 +290,7 @@ std::tuple<std::vector<int>, std::vector<std::pair<std::string,bool>>,
 zoltanGraphPartitionGridOnRoot(const CpGrid& cpgrid,
                                const std::vector<OpmWellType> * wells,
                                const double* transmissibilities,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                                const Communication<MPI_Comm>& cc,
-#else
-                               const CollectiveCommunication<MPI_Comm>& cc,
-#endif
                                EdgeWeightMethod edgeWeightsMethod,
                                int root,
                                const double zoltanImbalanceTol,
@@ -383,11 +377,7 @@ zoltanGraphPartitionGridOnRoot(const CpGrid& cpgrid,
 class ZoltanSerialPartitioner
 {
 public:
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
     using CommunicationType = Communication<MPI_Comm>;
-#else
-    using CommunicationType = CollectiveCommunication<MPI_Comm>;
-#endif
 
     ZoltanSerialPartitioner(const CpGrid& _cpgrid,
                             const std::vector<OpmWellType>* _wells,
@@ -582,11 +572,7 @@ std::tuple<std::vector<int>,
 zoltanSerialGraphPartitionGridOnRoot(const CpGrid& cpgrid,
                                      const std::vector<OpmWellType>* wells,
                                      const double* transmissibilities,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                                      const Dune::Communication<MPI_Comm>& cc,
-#else
-                                     const CollectiveCommunication<MPI_Comm>& cc,
-#endif
                                      EdgeWeightMethod edgeWeightsMethod,
                                      int root,
                                      const double zoltanImbalanceTol,
@@ -602,11 +588,7 @@ std::vector<int>
 zoltanGraphPartitionGridForJac(const CpGrid& cpgrid,
                                const std::vector<OpmWellType> * wells,
                                const double* transmissibilities,
-#if DUNE_VERSION_NEWER(DUNE_GRID, 2, 7)
                                const Dune::Communication<MPI_Comm>& cc,
-#else
-                               const CollectiveCommunication<MPI_Comm>& cc,
-#endif
                                EdgeWeightMethod edgeWeightsMethod, int root,
                                int numParts, const double zoltanImbalanceTol)
 {
@@ -618,4 +600,4 @@ zoltanGraphPartitionGridForJac(const CpGrid& cpgrid,
 
 } // namespace cpgrid
 } // namespace Dune
-#endif // HAVE_ZOLTAN
+#endif // defined(HAVE_ZOLTAN) && defined(HAVE_MPI)
