@@ -33,9 +33,11 @@
 */
 
 #include "config.h"
+
 #include <assert.h>
 #include <float.h>
 #include <math.h>
+#include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -44,7 +46,6 @@
 #include "uniquepoints.h"
 #include "facetopology.h"
 
-
 #define MIN(i,j) ((i)<(j) ? (i) : (j))
 #define MAX(i,j) ((i)>(j) ? (i) : (j))
 
@@ -669,12 +670,24 @@ get_zcorn_sign(int nx, int ny, int nz, const int *actnum,
 }
 
 
+/* ---------------------------------------------------------------------- */
+/* Compute (I,J,K) Cartesian coordinate of a single cell.
+ *
+ * @param[in] nx Number of Cartesian cells in model's X direction
+ * @param[in] ny Number of Cartesian cells in model's Y direction
+ * @param[in] nz Number of Cartesian cells in model's Z direction
+ * @param[in] c Linear Cartesian index, natural ordering, of model cell
+ * @param[out] i Cartesian I coordinate of cell \c c
+ * @param[out] j Cartesian J coordinate of cell \c c
+ * @param[out] k Cartesian K coordinate of cell \c c */
+/* ---------------------------------------------------------------------- */
 static void
 ind2sub(const size_t nx,
         const size_t ny,
         const size_t nz,
         size_t       c ,
         size_t *i, size_t *j, size_t *k)
+/* ---------------------------------------------------------------------- */
 {
     assert (c < (nx * ny * nz));
 
@@ -716,29 +729,262 @@ vert_size(const struct grdecl *in,
 
 
 /* ---------------------------------------------------------------------- */
-static int
-is_lefthanded(const struct grdecl *in, const double sign)
+/* Compute (X,Y,Z) coordinate of a single vertex on a pillar.
+ *
+ * @param[in] in Corner-point grid structure
+ * @param[in] pillar Pillar ID.  Zero to (nx+1)*(ny+1) - 1, inclusive.
+ * @param[in] z Z coordinate of vertex
+ * @param[out] coord Vertex coordinate.  Expected to point to the start of
+ *   an array of size at least 3. */
+/* ---------------------------------------------------------------------- */
+static void
+vertex_coord(const struct grdecl *in,
+             const size_t         pillar,
+             const double         z,
+             double              *coord)
 /* ---------------------------------------------------------------------- */
 {
-    int           active, searching;
-    size_t        nx, ny, nz, c;
-    size_t        origin, imax, jmax;
-    size_t        off[8];
-    double        dx[2], dy[2], dz, triple;
-    const double *pt_coord;
+    const double *top = &in->coord[6*pillar + 0];
+    const double *bot = &in->coord[6*pillar + 3]; /* == top + 3 */
+
+    /* Deem top and bottom pillar points coincident if Z coordinates along
+     * pillar differ by less than 1 micrometre */
+    const int coincide = fabs(top[2] - bot[2]) < 1.0e-6;
+
+    const double t = coincide
+        ? 0.0 /* coincide => vertical */
+        : (z - top[2]) / (bot[2] - top[2]); /* straight line */
+
+    coord[0] = (1.0 - t)*top[0] + t*bot[0];
+    coord[1] = (1.0 - t)*top[1] + t*bot[1];
+    coord[2] = z;
+}
+
+
+/* ------------------------------------------------------------------------ */
+/* Calculate cross product (I - O) x (J - O) when I and J are points on the
+ * I and J axes respectively, and O is the coordinate system origin.
+ *
+ * @param[in] origin Location of coordinate system's origin.  Expected to
+ *   point to the start of an array of size at least 3.
+ *
+ * @param[in] i_axis Location of point on coordinate system's I axis.
+ *   Expected to point to the start of an array of size at least 3.
+ *
+ * @param[in] j_axis Location of point on coordinate system's J axis.
+ *   Expected to point to the start of an array of size at least 3.
+ *
+ * @param[out] cross Resulting cross product.  Expected to point to the
+ *   start of an array of size at least 3.
+ */
+/* ------------------------------------------------------------------------ */
+static void
+bounding_box_cross_axes(const double *origin,
+                        const double *i_axis,
+                        const double *j_axis,
+                        double       *cross)
+{
+    cross[0] = (i_axis[1] - origin[1]) * (j_axis[2] - origin[2])
+        -      (i_axis[2] - origin[2]) * (j_axis[1] - origin[1]);
+
+    cross[1] = (i_axis[2] - origin[2]) * (j_axis[0] - origin[0])
+        -      (i_axis[0] - origin[0]) * (j_axis[2] - origin[2]);
+
+    cross[2] = (i_axis[0] - origin[0]) * (j_axis[1] - origin[1])
+        -      (i_axis[1] - origin[1]) * (j_axis[0] - origin[0]);
+}
+
+
+/* ------------------------------------------------------------------------ */
+/* Calculate triple product ((I - O) x (J - O)) . (K - O) when I,J,K are
+ * points on the I, J, and K axes respectively and O is the coordinate
+ * system origin.
+ *
+ * @param[in] origin Location of coordinate system's origin.  Expected to
+ *   point to the start of an array of size at least 3.
+ *
+ * @param[in] i_axis Location of point on coordinate system's I axis.
+ *   Expected to point to the start of an array of size at least 3.
+ *
+ * @param[in] j_axis Location of point on coordinate system's J axis.
+ *   Expected to point to the start of an array of size at least 3.
+ *
+ * @param[in] k_axis Location of point on coordinate system's K axis.
+ *   Expected to point to the start of an array of size at least 3.
+ *
+ * @return Triple product. */
+/* ------------------------------------------------------------------------ */
+static double
+bounding_box_triple_product(const double *origin,
+                            const double *i_axis,
+                            const double *j_axis,
+                            const double *k_axis)
+{
+    double cross[3];
+
+    bounding_box_cross_axes(origin, i_axis, j_axis, cross);
+
+    return cross[0]*(k_axis[0] - origin[0])
+        +  cross[1]*(k_axis[1] - origin[1])
+        +  cross[2]*(k_axis[2] - origin[2]);
+}
+
+
+/* ---------------------------------------------------------------------- */
+/* Known coordinate system types.  The 'Inconclusive' type is the
+ * default/failure type when we're not able to infer the actual type. */
+/* ---------------------------------------------------------------------- */
+enum CoordinateSystemType { Inconclusive, RightHanded, LeftHanded };
+/* ---------------------------------------------------------------------- */
+
+
+/* ---------------------------------------------------------------------- */
+/* Classify the coordinate system geometry type based on sign of triple
+ * product.  Positive value is right-handed, negative value is left-handed,
+ * and zero is inconclusive. */
+/* ---------------------------------------------------------------------- */
+static enum CoordinateSystemType
+classify_geometry(const double triple)
+/* ---------------------------------------------------------------------- */
+{
+    if (triple > 0.0) {
+        return RightHanded;
+    }
+    else if (triple < 0.0) {
+        return LeftHanded;
+    }
+    else {
+        return Inconclusive;
+    }
+}
+
+
+/* ---------------------------------------------------------------------- */
+/* Classify the coordinate system geometry type of a single active cell
+ * based on sign of cell's triple product
+ *
+ * @param[in] in Corner-point geometry.
+ * @param[in] i Cell's Cartesian I index.
+ * @param[in] j Cell's Cartesian J index.
+ * @param[in] k Cell's Cartesian K index.
+ * @param[in] sign ZCORN ordering sign.
+ * @param[in] off ZCORN vertex offsets for each of the cell's vertices.
+ * @return Coordinate system type.
+ */
+/* ---------------------------------------------------------------------- */
+static enum CoordinateSystemType
+get_cell_type(const struct grdecl *in,
+              const size_t         i,
+              const size_t         j,
+              const size_t         k,
+              const double         sign,
+              const size_t         off[8])
+/* ---------------------------------------------------------------------- */
+{
+    const size_t p0 = i + j*(in->dims[0] + 1);
+    const size_t pi = p0 + 1;
+    const size_t pj = p0 + in->dims[0] + 1;
+    const size_t io = 2*i + 2*in->dims[0]*(2*j + 2*in->dims[1]*2*k);
+
+    double triple, origin[3], I[3], J[3], K[3];
+
+    vertex_coord(in, p0, in->zcorn[io + off[0]], origin);
+    vertex_coord(in, pi, in->zcorn[io + off[1]], I);
+    vertex_coord(in, pj, in->zcorn[io + off[2]], J);
+    vertex_coord(in, p0, in->zcorn[io + off[4]], K);
+
+    triple = sign * bounding_box_triple_product(origin, I, J, K);
+
+    return classify_geometry(triple);
+}
+
+
+/* ---------------------------------------------------------------------- */
+/* Fallback coordinate system classification.  Computes the triple product
+ * of a bounding box around each active cell and forms a heuristic
+ * classification based on the number of cells of each type.  This
+ * classification is not backed by theory.
+ *
+ * @param[in] in Corner-point geometry.
+ * @param[in] sign ZCORN ordering sign.
+ * @param[in] off ZCORN vertex offsets for each of a cell's vertices.
+ * @return Coordinate system type.
+ */
+/* ---------------------------------------------------------------------- */
+static enum CoordinateSystemType
+coodinate_system_type_cell_criterion(const struct grdecl *in,
+                                     const double         sign,
+                                     const size_t         off[8])
+/* ---------------------------------------------------------------------- */
+{
+    size_t i, nx, j, ny, k, nz, c;
+    size_t ctype[3] = {0};
 
     nx = in->dims[0];
     ny = in->dims[1];
     nz = in->dims[2];
 
-    off[0] = 0;
-    off[1] = off[0] + 1;
-    off[2] = off[0] + (2 * nx);
-    off[3] = off[2] + 1;
-    off[4] = off[0] + ((2 * nx) * (2 * ny));
-    off[5] = off[4] + 1;
-    off[6] = off[4] + (2 * nx);
-    off[7] = off[6] + 1;
+    if (in->actnum == NULL) {
+        for (k = 0, c = 0; k < nz; k++) {
+            for (j = 0; j < ny; j++) {
+                for (i = 0; i < nx; i++, c++) {
+                    if (vert_size(in, c, off) > 0.0) {
+                        ++ ctype[get_cell_type(in, i, j, k, sign, off)];
+                    }
+                }
+            }
+        }
+    }
+    else {
+        for (k = 0, c = 0; k < nz; k++) {
+            for (j = 0; j < ny; j++) {
+                for (i = 0; i < nx; i++, c++) {
+                    if ((in->actnum[c] != 0) && (vert_size(in, c, off) > 0.0)) {
+                        ++ ctype[get_cell_type(in, i, j, k, sign, off)];
+                    }
+                }
+            }
+        }
+    }
+
+    if ((ctype[Inconclusive] > ctype[RightHanded]) &&
+        (ctype[Inconclusive] > ctype[LeftHanded]))
+    {
+        return Inconclusive;
+    }
+    else if (ctype[RightHanded] > ctype[LeftHanded]) {
+        return RightHanded;
+    }
+    else {
+        return LeftHanded;
+    }
+}
+
+
+/* ---------------------------------------------------------------------- */
+/* Primary coordinate system classification.  Computes the triple product of
+ * a bounding box around the model's geometry.
+ *
+ * @param[in] in Corner-point geometry.
+ * @param[in] sign ZCORN ordering sign.
+ * @param[in] off ZCORN vertex offsets for each of a cell's vertices.
+ * @return Coordinate system type. */
+/* ---------------------------------------------------------------------- */
+static enum CoordinateSystemType
+coordinate_system_type_model_bounding_box(const struct grdecl *in,
+                                          const double         sign,
+                                          const size_t         off[8])
+/* ---------------------------------------------------------------------- */
+{
+    int           active, searching;
+    size_t        nx, ny, nc, c;
+    size_t        origin, imax, jmax;
+    double        dx[2], dy[2], dz, triple;
+    const double *pt_coord;
+
+    nx = in->dims[0];
+    ny = in->dims[1];
+    nc = nx * ny * in->dims[2];
 
     pt_coord = in->coord;
 
@@ -763,7 +1009,7 @@ is_lefthanded(const struct grdecl *in, const double sign)
         searching = ! (active && (fabs(dz) > 0.0));
 
         c += 1;
-    } while (searching && (c < (nx * ny * nz)));
+    } while (searching && (c < nc));
 
     assert (! searching);       /* active && (fabs(dz) > 0) */
 
@@ -771,9 +1017,45 @@ is_lefthanded(const struct grdecl *in, const double sign)
      * from right-handed (>0) coordinate systems. */
     triple = sign * dz * (dx[0]*dy[1] - dx[1]*dy[0]);
 
-    assert (fabs(triple) > 0.0);
+    return classify_geometry(triple);
+}
 
-    return triple < 0.0;
+
+/* ---------------------------------------------------------------------- */
+/* Coordinate system classification.  Computes the triple product of a
+ * bounding box around the model's geometry.  Falls back to a per-cell
+ * heuristic criterion if initial test is inconclusive.
+ *
+ * @param[in] in Corner-point geometry.
+ * @param[in] sign ZCORN ordering sign.
+ * @return Coordinate system type. */
+/* ---------------------------------------------------------------------- */
+static enum CoordinateSystemType
+grid_coordinate_system_type(const struct grdecl *in, const double sign)
+/* ---------------------------------------------------------------------- */
+{
+    size_t                    nx, ny;
+    size_t                    off[8];
+    enum CoordinateSystemType coord_system_type;
+
+    nx = in->dims[0];
+    ny = in->dims[1];
+
+    off[0] = 0;
+    off[1] = off[0] + 1;
+    off[2] = off[0] + (2 * nx);
+    off[3] = off[2] + 1;
+    off[4] = off[0] + ((2 * nx) * (2 * ny));
+    off[5] = off[4] + 1;
+    off[6] = off[4] + (2 * nx);
+    off[7] = off[6] + 1;
+
+    coord_system_type =
+        coordinate_system_type_model_bounding_box(in, sign, off);
+
+    return (coord_system_type != Inconclusive)
+        ?  coord_system_type
+        :  coodinate_system_type_cell_criterion(in, sign, off);
 }
 
 
@@ -802,16 +1084,16 @@ reverse_face_nodes(struct processed_grid *out)
 }
 
 
-/*-----------------------------------------------------------------
-  Public interface
-*/
-void process_grdecl(const struct grdecl   *in,
-                    double                tolerance,
-                    const int*     is_aquifer_cell,
-                    struct processed_grid *out,
-                    int pinchActive)
+/* ----------------------------------------------------------------------
+ * Public interface
+ * ---------------------------------------------------------------------- */
+int process_grdecl(const struct grdecl   *in,
+                   double                 tolerance,
+                   const int             *is_aquifer_cell,
+                   struct processed_grid *out,
+                   int                    pinchActive)
 {
-    struct grdecl g;
+    struct grdecl g = {0};
 
     size_t i;
     int    sign, error, left_handed;
@@ -822,6 +1104,8 @@ void process_grdecl(const struct grdecl   *in,
 
     double *zcorn;
 
+    enum CoordinateSystemType coord_sys_type;
+
     const size_t BIGNUM = 64;
     const int    nx = in->dims[0];
     const int    ny = in->dims[1];
@@ -834,7 +1118,12 @@ void process_grdecl(const struct grdecl   *in,
     int    *intersections;
 
 
+    sign = get_zcorn_sign(nx, ny, nz, in->actnum, in->zcorn, &error);
+    coord_sys_type = grid_coordinate_system_type(in, sign);
 
+    if (error || (coord_sys_type == Inconclusive)) {
+        return 0;
+    }
 
     /* -----------------------------------------------------------------*/
     /* Initialize output structure:
@@ -861,7 +1150,14 @@ void process_grdecl(const struct grdecl   *in,
     out->node_coordinates = NULL;
     out->local_cell_index = malloc(nc * sizeof *out->local_cell_index);
 
-
+    if ((out->face_neighbors   == NULL) ||
+        (out->face_nodes       == NULL) ||
+        (out->face_ptr         == NULL) ||
+        (out->face_tag         == NULL) ||
+        (out->local_cell_index == NULL))
+    {
+        return 0;
+    }
 
     /* Do actual work here:*/
 
@@ -878,27 +1174,38 @@ void process_grdecl(const struct grdecl   *in,
     g.dims[1] = in->dims[1];
     g.dims[2] = in->dims[2];
 
-    actnum    = malloc (nc *     sizeof *actnum);
-    g.actnum  = copy_and_permute_actnum(nx, ny, nz, in->actnum, actnum);
+    actnum = malloc(nc * sizeof *actnum);
+    if (actnum == NULL) {
+        return 0;
+    }
 
-    zcorn     = malloc (nc * 8 * sizeof *zcorn);
-    sign      = get_zcorn_sign(nx, ny, nz, in->actnum, in->zcorn, &error);
-    g.zcorn   = copy_and_permute_zcorn(nx, ny, nz, in->zcorn, sign, zcorn);
+    g.actnum = copy_and_permute_actnum(nx, ny, nz, in->actnum, actnum);
 
-    g.coord   = in->coord;
+    zcorn = malloc (nc * 8 * sizeof *zcorn);
+    if (zcorn == NULL) {
+        free(actnum);
+        return 0;
+    }
 
+    g.zcorn = copy_and_permute_zcorn(nx, ny, nz, in->zcorn, sign, zcorn);
+    g.coord = in->coord;
 
     /* allocate space for cornerpoint numbers plus INT_MIN (INT_MAX)
      * padding */
     plist = malloc(8 * (nc + ((size_t)nx)*((size_t)ny)) * sizeof *plist);
+    if (plist == NULL) {
+        free(zcorn);
+        free(actnum);
+        return 0;
+    }
 
     finduniquepoints(&g, plist, tolerance, out);
 
-    free (zcorn);
-    free (actnum);
+    free(zcorn);  zcorn  = NULL;
+    free(actnum); actnum = NULL;
 
     /* Determine if coordinate system is left handed or not. */
-    left_handed = is_lefthanded(in, sign);
+    left_handed = coord_sys_type == LeftHanded;
     if (left_handed) {
         /* Reflect Y coordinates about XZ plane to create right-handed
          * coordinate system whilst processing intersections. */
@@ -912,26 +1219,34 @@ void process_grdecl(const struct grdecl   *in,
 
     /* internal */
     work = malloc(2 * ((size_t) (2*nz + 2)) * sizeof *work);
-    for(i = 0; i < ((size_t)4) * (nz + 1); ++i) { work[i] = -1; }
+    if (work == NULL) {
+        free(plist);
+        return 0;
+    }
+
+    for (i = 0; i < ((size_t)4) * (nz + 1); ++i) { work[i] = -1; }
 
     /* internal array to store intersections */
     intersections = malloc(BIGNUM* sizeof(*intersections));
-
-
+    if (intersections == NULL) {
+        free(plist);
+        free(work);
+        return 0;
+    }
 
     process_vertical_faces   (0, &intersections, plist, work, out);
     process_vertical_faces   (1, &intersections, plist, work, out);
     process_horizontal_faces (   &intersections, plist, is_aquifer_cell, out, pinchActive);
 
-    free (plist);
-    free (work);
+    free(work);   work  = NULL;
+    free(plist);  plist = NULL;
 
     /* -----------------------------------------------------------------*/
     /* (re)allocate space for and compute coordinates of nodes that
      * arise from intersecting cells (faults) */
     compute_intersection_coordinates(intersections, out);
 
-    free (intersections);
+    free(intersections);  intersections = NULL;
 
     /* -----------------------------------------------------------------*/
     /* Enumerate compressed cells:
@@ -940,6 +1255,10 @@ void process_grdecl(const struct grdecl   *in,
        lexicographically ordered, used to remap out->face_neighbors
     */
     global_cell_index = malloc(nc * sizeof *global_cell_index);
+    if (global_cell_index == NULL) {
+        return 0;
+    }
+
     cellnum = 0;
     for (i = 0; i < nc; ++i) {
         if (out->local_cell_index[i] != -1) {
@@ -957,7 +1276,6 @@ void process_grdecl(const struct grdecl   *in,
         }
     }
 
-
     free(out->local_cell_index);
     out->local_cell_index = global_cell_index;
 
@@ -985,10 +1303,13 @@ void process_grdecl(const struct grdecl   *in,
     if (left_handed ^ (sign == -1)) {
         reverse_face_nodes(out);
     }
+
+    return 1;
 }
 
-/*-------------------------------------------------------*/
+/* ---------------------------------------------------------------------- */
 void free_processed_grid(struct processed_grid *g)
+/* ---------------------------------------------------------------------- */
 {
     if( g ){
         free ( g->face_nodes       );
