@@ -8,8 +8,10 @@
 #include <opm/grid/CpGrid.hpp>
 #include <opm/grid/common/CommunicationUtils.hpp>
 #include <opm/input/eclipse/Deck/Deck.hpp>
+#include <opm/input/eclipse/EclipseState/Grid/EclipseGrid.hpp>
 #include <opm/input/eclipse/Parser/ErrorGuard.hpp>
 #include <opm/input/eclipse/Parser/Parser.hpp>
+#include <opm/input/eclipse/Parser/ParseContext.hpp>
 #if HAVE_MPI
 class MPIError {
 public:
