@@ -66,11 +66,7 @@ public:
           dist_cell_ids_(dist_cell_ids)
     {}
     typedef int DataType;
-#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 7)
     bool fixedSize(int /*dim*/, int /*codim*/)
-#else
-    bool fixedsize(int /*dim*/, int /*codim*/)
-#endif
     {
         return true;
     }
@@ -165,11 +161,7 @@ public:
 
     typedef int DataType;
 
-#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 7)
     bool fixedSize(int /*dim*/, int /*codim*/)
-#else
-    bool fixedsize(int /*dim*/, int /*codim*/)
-#endif
     {
         return true;
     }
@@ -278,11 +270,7 @@ class DummyDataHandle
 {
 public:
     typedef double DataType;
-#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 7)
     bool fixedSize(int /*dim*/, int /*codim*/)
-#else
-    bool fixedsize(int /*dim*/, int /*codim*/)
-#endif
     {
         return true;
     }
@@ -323,11 +311,7 @@ public:
     {}
 
     typedef int DataType;
-#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 7)
     bool fixedSize(int /*dim*/, int /*codim*/)
-#else
-    bool fixedsize(int /*dim*/, int /*codim*/)
-#endif
     {
         return true;
     }
